#include "Frame.h"

/** Default ctor.
* This is equivalent to creating a field with no pixels.
*/
Frame::Frame() : posit()
{
	nx =ny =0 ;
}

/** Ctor of a frame anchored at the origin with known dimensions.
* The frame ranges from 0<=x<dimx and 0<=y<dimy.
* @param dimx The number of pixels in the x direction
* @param dimy The number of pixels in the y direction
*/
Frame::Frame(int dimx, int dimy) 
	: nx(dimx), ny(dimy), posit()
{
}

/** General ctor of a frame anchored at the origin with known dimensions.
* @param dimx The number of pixels in the x direction
* @param dimy The number of pixels in the y direction
* @param pt The position of the llx and lly point of the frame in a larger scope.
*/
Frame::Frame(int dimx, int dimy, const Point &pt) 
	: nx(dimx), ny(dimy), posit(pt)
{
}


/** Test if the pixel is inside this window.
* @param x The x-coordinate of the pixel
* @param y The y-coordinate of the pixel
* @return true if the pixel specified is one of the pixels inside this array.
* x coordinates equal to posit.x and y coordinates equatl to posit.y are
* considered to be outside for this purpose.
*/
bool Frame::covers(int x, int y) const
{
	/* The pixel is inside this array if 0<=x <ny and 0<=y < ny
	*/
	return ( x >= posit.x && x < posit.x+nx && y >= posit.y && y < posit.y+ny) ;
}

/** Test if this is entirely inside another frame.
* @param windo The reference array .
* @return True if this array is completely covered by windo; false if
*   there is no or only partial overlap.
*/
bool Frame::inside(const Frame & windo) const
{
	/* To test whether this is completely covered by windo, is suffices
	* to check whether all four edges of this are within the
	* coordinate intervals set by windo.
	* This here has bounding box [posit.x,posit.y,posit.x+nx,posit.y+ny],
	* the other window has bounding box [windo.posit.x, windo.posit.y,..,..].
	*/
	return
		windo.covers(posit.x, posit.y)
		&& windo.covers(posit.x, posit.y+ny-1)
		&& windo.covers(posit.x+nx-1, posit.y)
		&& windo.covers(posit.x+nx-1, posit.y+ny-1) ;
}



/** The integer area in units of pixels squared.
*/
int Frame::area() const
{
	return nx*ny ;
}

/** Display the bounding box of a frame.
* @param os The output stream to print to
* @param wd The frame to be displayed.
* @return The updated output stream
*/
std::ostream & operator<<(std::ostream & os, const Frame & wd)
{
	os << "[" << wd.posit.x <<"," << wd.posit.y << "," 
		<< wd.posit.x+wd.nx << "," << wd.posit.y+wd.ny << "] "  ;
	return os ;
}
