#include <complex>
#include "RecPixSupArr.h"

/** Ctor with no window.
*/
template <class TYP> RecPixSupArr<TYP>::RecPixSupArr() : vals()
{
}





/** Set all values in all windows.
* @param[in] fval An array with as many values as the area of all subwindows.
* @param[in] fs The geometry information about the layout of the subwindows.
*/
template <class TYP> template <class TYPV> void RecPixSupArr<TYP>::set(const TYPV *fval, const FrameSet & fs)
{
	/* Loop over the frames in this frame set
	*/
	for(int w=0 ; w < fs.wd.size() ; w++)
	{
		/* Prepare a subwindow pixel frame
		*/
		RecPixArr<TYP> r( fs.wd[w].nx, fs.wd[w].ny) ;

		/* Copy the associated chunk of fval[] to the frame.
		*/
		r.set(fval + fs.areas[w]);

		/* Copy the subwindow into this modes frame set
		*/
		vals.push_back(r) ; 
	}
}

/** Read the value of a pixel.
* @param windo The 0-based subwindow number
* @param x The 0-based index in the x direction of that subwindow
* @param y The 0-based index in the y direction of that subwindow
* @return The current value at the specified address.
* Requests to read a pixel outside the region defined by the array return zero.
*/
template <class TYP> TYP RecPixSupArr<TYP>::at(int x, int y, const FrameSet & fs) const
{
	/* search for the window number to hold the value
	*/
	const int win = fs.covers(x,y) ;

	/* forward the call to the specific window, relocated into its local coordinates
	*/
	if ( win >= 0)
		return vals[win].at(x -fs.wd[win].posit.x,y -fs.wd[win].posit.y) ;
	else
		return 0. ;
}

/** Convert the values of all pixels to a valarray.
* The representation returned is linearized with successive row[0], row[1] etc
* @param[in] fs The structural information on size and placement of the windows.
* @return The valarray representation with the pixel values.
* Pixels in the interstitial regions (outside any windows) are filled with zeros.
*/
template <class TYP> std::valarray<TYP> RecPixSupArr<TYP>::toValarray(const FrameSet & fs) const
{
	/* The dimensions of the outer frame circumscribing all windows.
	*/
	int bbox[4] ;
	fs.boundbox(bbox) ;

	/* count total pixels, including the interstitial areas.
	*/
	int dim[2] ;
	dim[0] = bbox[2]-bbox[0] ;
	dim[1] = bbox[3]-bbox[1] ;

	/* The output stream of all pixel values. Row 0 first.
	*/
	/* Initialize the output with zeros and then fill in the 
	* values fromthe individual frames.
	*/
	std::valarray<TYP> strea(static_cast<TYP>(0),dim[0]*dim[1]) ;
	for(int w=0 ; w < vals.size() ; w++)
	{
		/* The origin of window w is fs.wd[w].posit, and by subraction
		* of bbox[] we obtain the relative position in the strea .
		*/
		int offs[2] ;
		offs[0] = fs.wd[w].posit.x - bbox[0] ;
		offs[1] = fs.wd[w].posit.y - bbox[1] ;
		for(int y=0 ; y < vals[w].ny ; y++)
			for(int x=0 ; x < vals[w].nx ; x++)
				strea[(y+offs[1])*dim[0]+x+offs[0]] = vals[w].at(x,y) ;
	}
	
		
	return strea ;
}

/* Explicit template instantiation to avoid link time errors by gcc
*/
template class RecPixSupArr<double>;
template class RecPixSupArr<float>;
template class RecPixSupArr<int>;
template class RecPixSupArr<short>;
template class RecPixSupArr<long>;
template void RecPixSupArr<double>::set(const double *fval, const FrameSet & fs) ;
template void RecPixSupArr<float>::set(const double *fval, const FrameSet & fs) ;

