#include <cmath>
#include <string>
#include <iostream>
#include <typeinfo>

/* The GSL (Gnu scientific library) of www.gnu.org/software/gsl
*/
#include <gsl/gsl_sf.h>

#include "VKarCorr.h"

#ifdef HAVE_CCFITS
#include <CCfits>
#endif

#if 0

/* The inclusion of the following two files leads to unresolvable name
* clashes between the complex struct defined in the library and the standard
* C++ complex class that is used through this project. So we expand
* the contents to the single line serving our needs and hope that in some
* future versions that problem will be avoidable.
*/
extern "C" {
/** The f2c header file f2c.h.
* A prerequisite to include clapack.h below.
* This is obtained from <a href="http://www.netlib.org/f2c">www.netlib.org/f2c</a>.
*/
#include <f2c.h>
}

extern "C" {
/** The clapack header file \c clapack.h 3.2.1.
* This is obtained from <a href="http://www.netlib.org/clapack">www.netlib.org/clapack</a>.
*/
#include <clapack.h>
}
#else /* 0 */

/* Interface to www.netlib.org/clapack 3.2.1
*/
extern "C" {
	int dsyev_(char *jobz, char *uplo, long int *n, double *a, 
	 long int *lda, double *w, double *work, long int *lwork, 
	long int *info);
	int dlarnv_(long int* idist, long int * iseed, long int *n, double *x) ;
}

#endif /* 0 */

#ifndef M_PI
/* In case cmath does not define Pi, the value is made available explictly */
#define M_PI 3.14159265358979323846264338328
#endif


/** Ctor with the basic physical characteristics.
* @param rho0 The Fried parameter [m]
* @param ouscale The outer scale length [m]
*   The default is zero, which generates the Kolmogorov spectrum.
* @param kolgamma The unitless exponent of the power spectrum.
*   The default is 2/3, which yields the Kolmogorov standard slope.
*/
template <class TYP> VKarCorr<TYP>::VKarCorr(const double rho0, const double ouscale, const double kolgamma) 
	: rFried(rho0), outscal(ouscale), kolg_exp(kolgamma), modes(), screens()
{
	/* Compute two constants to avoid frequent recomputation with each
	* call of corr() in the sequel
	*/
	const double g = 0.5*(1.+kolg_exp) ;

	/* The value of 2*cphi is roughly 6.8838 at kolg_exp=2/3.
	*/
	cphi = pow(4.*gsl_sf_gamma(1.+1./g),g) ;
	Cfact = 2.*cphi/gsl_sf_gamma(-g) ;

	outscalwn = (outscal >0.)? 2.*M_PI/outscal :0.;
}


/** Evaluate the correlation function for a distance between two
* points.
* The applicable formula of the structure function is
* \latexonly
* \cite[eq (9)]{MatharArxiv0911}
* \endlatexonly
* \f[
* D_\varphi = 
* -(P/r_0)^{1+\gamma} \frac{1}{\Gamma(-1/2-\gamma/2)(\pi P/L_0)^{1+\gamma}}
* \left[\Gamma(1/2+\gamma/2)-2(\pi P/L_0)^{1/2+\gamma/2}K_{1/2+\gamma/2}(2\pi P/L_0)\right]
* \f]
* with Fried radius \f$r_0\f$, outer scale \f$L_0\f$, Kolmogorov exponent \f$\gamma\f$.
* The limit of Kolmogorov turbulence is
* \latexonly
* \cite{MatharWRCM20,MatharArxiv0705b}
* \endlatexonly
* \f[
* D_\varphi = 2 c_\varphi (P/r_0)^{1+\gamma}
* \f]
* with \f$g(\gamma) =\sqrt{\pi}\Gamma(-1/2-\gamma/2)/\Gamma(-\gamma/2)\f$.
* The value returned is \f$-1/2\f$ of this, where sign change and factor mark the
* transition to the correlation function.
* @param P The distance between the two points in the pupil [m]
* @return The value of the bivariate phase correlation  function at that distance.
*/
template <class TYP> double VKarCorr<TYP>::corr(const double P) 
{
	/* Dphi = kbar^2 *h *Cn^2 * P^(1+gamma)*sqrt(pi)/Gamma(-gamma/2)/(Pi*P*f0)^(1+gamma)
	*  [Gamma(1/2+gamma/2)-2*(pi*P*f0)^(1/2+gamma/2)*K_{1/2+gamma/2}(2*pi*P*f0)]
	*/

	if ( outscalwn <=0.)
	{
		/* case of pure Kolmogorov, infinit outer scale */
		return -cphi*pow(P/rFried,1.0+kolg_exp) ;
	}
	else
	{
		if ( P == 0.)
			return 0. ;

		/* (1+gamma)/2 of the Kolmogorov scale
		*/
		const double gammprime = 0.5*(1.0+kolg_exp) ;

		/* Start with the unit-less constant
		*/
		double C = -2.*pow(0.5*P*outscalwn,gammprime)*gsl_sf_bessel_Knu( gammprime,P*outscalwn) ;
		C += gsl_sf_gamma(gammprime) ;
		return  0.5*Cfact * C * pow(P/rFried,1.+kolg_exp )  / pow(0.5*P*outscalwn,1.+kolg_exp) ;
	}
}



/** An ASCII representation of all modes and screens.
* The Fried parameter and outer scale are printed comment style, i.e.,
* preceded by a <tt>#</tt>, then the modes, then the phase screens.
* The format is for each mode and each screen a description of
* parameters in comments (ie, in lines starting with the sharp). Each line is
* the x and y value of the pixel in meters followed by the phase value of the mode
* in radians. Note that the modes are not yet multiplied by the (square root of) the
* eigenvalue. After each pixel row follows an empty lines. Two empty lines separate
* the modes and screens. This format is exactly the one needed to display the values
* scanning with gnuplot commands through the data,
* <tt> gnuplot</tt>
*
* <tt> gnuplot> set style data lines</tt>
*
* <tt> gnuplot> set hidden3d</tt>
*
* <tt> gnuplot> splot "out.asc" index 0</tt>
*
* <tt> gnuplot> splot "out.asc" index 1</tt>
*
* <tt> gnuplot> splot "out.asc" index 2</tt>
*
* <tt> gnuplot> quit</tt>
*
* @param os The output stream to be written to.
* @param fs The frame set description with the geometric paramters.
*/
template <class TYP> void VKarCorr<TYP>::plot(std::ostream & os, const FrameSet & fs)
{
	os << "# Fried parameter " << rFried << " m\n" ;
	os << "# outer scale " ;
	if ( outscal > 0)
		os << outscal << " m\n" ;
	else
		os << "infinity\n" ;

	for(int m= modes.size()-1 ; m >= 0 ; m--)
	{
		os << "# mode " << m << std::endl ;
		modes[m].plot(os,fs) ;
		os << "\n\n" ;
	}
	for(int s=0 ; s < screens.size() ; s++)
	{
		os << "# screen " << s << std::endl ;
		screens[s].plot(os,fs) ;
		os << "\n\n" ;
	}
}

#ifdef HAVE_CCFITS
/** Convert the modes and screens to a FITS file.
* It has been a design choice by the CCfits maintainers for version 2.2 to
* disallow copy construction on return, so we return a pointer to the local
* object rather than the object itself.
* The keywords in the primary header are
*   - <tt>RFRIED</tt> The Fried radius in units of meter
*   - <tt>OUTSCA</tt> The outer scale in meter. A negative value means \f$\infty\f$
*   - <tt>MODES</tt> The number of Karhunen-Lo\`eve modes. Each plane in the FITS cube is one mode. To reach to the phase screens, this number of planes must be skipped.
*   - <tt>SCREENS</tt> The number of phase screens. This is the number planes in the FITS cube minus the number of modes.
*   - <tt>KOLGEXP</tt> The Kolmogorov scale exponent.
*   - <tt>FRLLX</tt> Keywords that start with FR describe the frames in units of pixel coordinates.
*           <tt>LLX</tt>, <tt>LLY</tt>, <tt>URX</tt> and <tt>URY</tt> refer to the
*           lower left x, lower left y, upper right x and upper right y coordinate
*           of a bounding box of a rectangular region.
*
*  Other keywords are strictly following the FITS standard and relate
*  the pixel units to meters.
*
*  The value reported with each pixel of the modes is already multiplied
*  by the associated eigenvalue. That is, if the eigenvalue equation is
*  \f[ \int d^2r' C_\varphi(r,r') K_j(r') = {\cal B}_j^2 K_j(r)\f],
*  the values in the file are \f${\cal B}_j K_j(r)\f$ for mode \f$j\f$.
*  (This is a difference to the format of the standard ASCII output.)
*
*  Values outside the stripes and not covered by the modes are set to <tt>NULL</tt>.
* @param fname The file name of the FITS file that is to be generated.
* @param fs The information on the layout of the subwindows.
*/
template <class TYP> CCfits::FITS * VKarCorr<TYP>::toFITS(const std::string & fname,const FrameSet & fs) const
{
	/* Fits 3d cube
	*/
	long naxis =3 ;
	long naxes[3] ;
	naxes[2] = modes.size()+screens.size() ;

	int bbox[4] ;
	fs.boundbox(bbox) ;
	naxes[0] = bbox[2]-bbox[0] ;
	naxes[1] = bbox[3]-bbox[1] ;

	/* pixel count (including interstitials) of one layer
	*/
	long ar = naxes[0]*naxes[1] ;

	TYP nullVal = 0. ;
	CCfits::FITS *fout = new CCfits::FITS(fname, (typeid(TYP) == typeid(double)) ? DOUBLE_IMG: FLOAT_IMG,naxis,naxes) ;

	CCfits::PHDU & h = fout -> pHDU() ;
	/* push the modes followed by the screens into the fits cube.
	*/
	for(int m= modes.size()-1 ; m >= 0 ; m--)
	{
		std::valarray<TYP> vm = modes[m].toValarray(fs) ;

		/* Store the modes scaled with the eigenvalue, because with the
		* current simple format this information is not stored at an extra position
		* in the file, but it is needed to generate other information afterwards.
		*/
		vm *= modes[m].eval ;

		/* Indices in cfitsio are 1-based 
		*/
		h.write(1+(long)(m*ar),ar,vm, &nullVal) ;
	}
	for(int s=0 ; s < screens.size() ; s++)
	{
		std::valarray<TYP> vm = screens[s].toValarray(fs) ;
		h.write(1+long((s+modes.size())*ar),ar,vm, &nullVal) ;
	}
	h.addKey("OUTSCA",outscal,"[m] outer scale, negative value means infinite") ;
	h.addKey("RFRIED",rFried,"[m] Fried radius r0") ;
	h.addKey("MODES",modes.size(),"[] modes are image planes 1..this value") ;
	h.addKey("SCREENS",screens.size(),"[] all remaining MODES+1..final image plane") ;
	h.addKey("KOLGEXP",kolg_exp,"[] Kolmogorov exponent") ;
	h.addKey("BUNIT","rad","phases measured in radians") ;
	h.addKey("CTYP1","x","name of one edge Cartesian coordinate") ;
	h.addKey("CTYP2","y","name of one edge Cartesian coordinate") ;
	h.addKey("CTYP3","#","third cube axis is mode or screen number") ;
	h.addKey("CRPIX1",1.,"reference pixel x") ;
	h.addKey("CRPIX2",1.,"reference pixel y") ;
	/* ensure that modes[0] exists to get the pixel scale
	*/
	if ( modes.size() )
	{
		h.addKey("CRVAL1",bbox[0]*modes[0].pixscal,"position of reference pixel x") ;
		h.addKey("CRVAL2",bbox[1]*modes[0].pixscal,"position of reference pixel y") ;
		h.addKey("CDELT1",modes[0].pixscal,"pixel scale along x") ;
		h.addKey("CDELT2",modes[0].pixscal,"pixel scale along y") ;
	}
	h.addKey("FRAMES",fs.wd.size(),"Number nonoverlapping frames") ;
	for(int f =0 ; f < fs.wd.size() ; f++)
	{
		std::ostringstream kw[4] ;
		kw[0] << "FRLLX"  << f+1 ;
		h.addKey(kw[0].str(), 1+fs.wd[f].posit.x-bbox[0],"1-based pixel lower left x") ;
		kw[1] << "FRLLy"  << f+1 ;
		h.addKey(kw[1].str(), 1+fs.wd[f].posit.y-bbox[1],"1-based pixel lower left y") ;
		kw[2] << "FRURX"  << f+1 ;
		h.addKey(kw[2].str(), 1+fs.wd[f].posit.x+fs.wd[f].nx-1-bbox[0],"1-based pixel upper right x") ;
		kw[3] << "FRURY"  << f+1 ;
		h.addKey(kw[3].str(), 1+fs.wd[f].posit.y+fs.wd[f].ny-1-bbox[1],"1-based pixel upper right y") ;
	}
	h.addKey("CREATOR",__FILE__,"Richard J. Mathar www.strw.leidenuniv.nl") ;
	h.writeComment("Wav Rand Compl Med 20 (2010) 23, stripe mode examples") ;
	h.writeComment("arXiv:0911.4710, Eq (9) with vK spectrum") ;
	h.writeDate() ;
	return fout ;
}

#endif /* HAVE_CCFITS */

/** Generate the Karhunen-Loeve modes.
* This populates the array of modes that is part of the object.
* @param[in] fs The parameters of the finite elements (pixels) as a number of frames.
* @param[in] pixscal The length and width of each pixel in units of meter.
*/
template <class TYP> void VKarCorr<TYP>::genmods(const FrameSet & fs, const double pixscal)
{
	/* The eigenvalue problem has matrix orders proportional to the area.
	*/
	int dim = fs.area() ;

	/* The correlation matrix to be diagonalized. This is always done
	* in double precision independent of the TYP of the C++ class.
	*/
	double * corrM = new double[dim*dim] ;

	/* The positions of the two points on the phase screen in natural units.
	*/
	double xy[2][2] ;

	int r3idx[3] ;
	for(int ridx=0 ; ridx < dim ; ridx++)
	{
		/* Relocate the row index into an information related
		* to the entire frame set's origin.
		*/
		fs.loc(ridx, r3idx) ;

		/* Add the origin of the corresponding window and the relative position
		* inside the window and scale with the pixel length scale.
		*/
		xy[0][0] = pixscal* (fs.wd[r3idx[0]].posit.x + r3idx[1]) ;
		xy[0][1] = pixscal* (fs.wd[r3idx[0]].posit.y + r3idx[2]) ;

		/* Matrix is symmetric because the correlation is symmetric. Fill only one part.
		*/
		for(int cidx=ridx ; cidx < dim ; cidx++)
		{
			/* The same computation as with the row index repeated for the column index.
			*/
			fs.loc(cidx, r3idx) ;
			xy[1][0] = pixscal* (fs.wd[r3idx[0]].posit.x + r3idx[1]) ;
			xy[1][1] = pixscal* (fs.wd[r3idx[0]].posit.y + r3idx[2]) ;
			corrM[ridx*dim+cidx] = corrM[cidx*dim+ridx] 
				= pixscal*pixscal*corr( hypot( xy[0][0]-xy[1][0], xy[0][1]-xy[1][1]) );

		}
	}

	/* Interface with the LAPACK library.
	*/
	char jobz[2] = "V" ;
	char uplo[2] = "U" ;
	long int lda = dim ;
	long int lwork = 3*dim ;
	long int info ;
	double *work = new double[lwork] ;
	double *lambda = new double[dim] ;

	/* The finite element version of the integral equation with symmetric matrix
	* is solved by a call to LAPACK's dsyev().
	*/
	dsyev_(jobz,uplo,&lda,corrM,&lda,lambda,work,&lwork,&info) ;
        if ( info)
                std::cerr << __FILE__ << " " << __LINE__ << " dsyev info " << info << std::endl ;
        else
        {
		/* Invalidate any phase screens that might have been left over
		*/
		modes.clear() ;

		/* TODO: normalize the modes such that the integral over the area is 1.
		* Since the integral is the sum times the pixel area, ensure that sum(mode[pix]^2)*pixscal^2=1
		*/

		/* Each column of the resulting array represents one mode.
		*/
		for(int cidx=0 ; cidx < dim ; cidx++)
		{
			/* Start with an empty mode
			*/
			RecPixSupArr<TYP> mod ;

			/* Copy the associated chunk corrM[cidx*dim.. (cidx+1)*dim-1] to the frames.
			*/
			mod.set(corrM+cidx*dim,fs) ;

			/* Declare this to be a phase screen by adding pixel scale and
			* mode eigenvalue information.
			*/
			PhaScr<TYP> mc(mod,pixscal) ;
			if (lambda[cidx] >0.)
			{
				mc.eval = sqrt(lambda[cidx]) ;
	
				/* Save the mode
				*/
				modes.push_back(mc) ;
			}
		}

		/* Invalidate any phase screens that might have been left over
		*/
		screens.clear() ;
        }

	/* Delete temporary work areas allocated above.
	*/
        delete [] lambda ;
        delete [] work ;
	delete [] corrM ;
}

/** Generate randomized phase screens.
* This populates the array of screens that is part of the object.
* @param[in] Nscr The number of screens to be generated
* @param[in] seed The seed of the random number generator.
*           Only the lowest 12 bits are actually used.
*/
template <class TYP> void VKarCorr<TYP>::genscr(const int Nscr, const int seed, const FrameSet &fs)
{
	/* initialize the clapack random number generator 
	*/
	long int lapsee[4] ;
	lapsee[0]=lapsee[1]=lapsee[2]=0 ;
        lapsee[3] = (seed & 4095) | 1 ;

	/* Loop over the screens as often as requested
	*/
	for(int s=0 ; s < Nscr ;s++)
	{
		/* for one screen one needs as my random numbers as modes */
		long int iN = modes.size() ;
		double * rstd = new double[modes.size()] ;

		/* A call to dlarnv() of the LAPACK library produces \c iN Gaussian random numbers
	        * with mean 0 and standard deviation 1.
	        */
		long int idist = 3;
	        dlarnv_(&idist,lapsee,&iN,rstd) ;

		/* Empty initial screen
		*/
		RecPixSupArr<TYP> scrArr ;

		/* Loop over all subwindows */
		for(int w=0 ; w < fs.wd.size() ; w++)
		{
			/* Start with a window all values zeroed
			*/
			std::valarray<TYP> rscr(static_cast<TYP>(0), fs.wd[w].area() ) ;

			/* Add all modes multiplied by eigenvalue and random number inside that window.
			*/
			for(int m=0; m < iN; m++)
				rscr += static_cast<TYP>(rstd[m] * modes[m].eval) * modes[m].vals[w].val ;
			
			/* Create a new RecPixArr and store the screen just created
			*/
			RecPixArr<TYP> wscr(fs.wd[w].nx, fs.wd[w].ny, rscr) ;

			/* Window w completed. Store it in the super array.
			*/
			scrArr.vals.push_back(wscr) ;
		}

		/* Transform the information to a PhaScre and store it.
		*/
		PhaScr<TYP> onescr(scrArr, modes[0].pixscal) ;
		screens.push_back(onescr) ;

		delete [] rstd ;
	}
}

/* Explicit template instantiation; seems to be needed with the gcc compiler even with -frepo
*/
template class VKarCorr<double> ;
template class VKarCorr<float> ;
