#include <iostream>
#include <vector>
#include <cstdlib>
#include <cmath>
#include "FrameSet.h"
#include "VKarCorr.h"
/** VKarPhase use and command line options.
* 
* <tt> VKarPhase [-w width] [-l length] [-P pixnum] [-N screenNo] [-r Fried] [-L outerscl] [-f out.fits] [-s seed] x0 y0 [x1 y1...] > out.asc </tt>
*
* The comand line options in decreasing order of importance are
*   - <tt>-w</tt> The width of each individual stripe in meters.
*      The default is 1.8.
*      Note that this the actual coverage will differ from this number
*      because it will be readjusted to a multiple of the pixel scale.
*   -   <tt>-l</tt> The length of each individual stripe in meters.
*      The default is 1.8.
*      Note that this the actual coverage will differ from this number
*      because it will be readjusted to a multiple of the pixel scale.
*   -   <tt>-P</tt> The estimate of the pixel count in each of the screens and modes.
*      The default is 100, which is probably too small for most applications.
*      The time of computation grows with some power of this number, so instead of
*      setting this value to arbitrarily large numbers with no chance to compute
*      in any finite time it is advised to increase this parameter by small
*      factors near 1.5 or 2 to gain exprience of typical run times.
*      The number refers to the total of all pixels (square elements) in the
*      union of the stripes; the interstitial region in between the stripes
*      is not included in this count.
*   -   <tt>-N</tt> The number of phase screens generated.
*      The default is 100.
*   -   <tt>r</tt> The Fried radius in units of meters.
*      The default is 0.15.
*      This is the scaling factor for the seeing and should be set to a value
*      that is reasonable for the observing wavelength of the electromagnetic spectrum.
*   -   <tt>-L</tt> The outer scale radius of the von-Karman spectrum.
*      The default is -1.
*      Negative values trigger the use of the Kolmogorov limit (infinite outer scale).
*   -   <tt>-f</tt> The name of the FITS file to be created.
*      There is no default. 
*      Note that existing files will not be overwritten. If this file exists
*      at run time, an associated error message from CCfits ensues.
*   -   <tt>-s</tt> The seed of the pseudo random number generator.
*      The default is 0. To generate additional phase screens after
*      having done this before, the seed must be changed, because this is
*      indeed working with a pseudo random number generator.
*   -   <tt>-g</tt> The scale parameter in the exponent of the 3D structure function.
*      The default is 2/3, the usual Kolmogorov power law.
*
* The command line parameters follow the command line options and are pairs of
* floating point numbers indicating the lower left coordinates in x and y of the 
* placement of the stripes. The number of stripes is half of the number of these
* parameters.
*
* If the stripes defined this way overlap by at least one pixel, the number
* of stripes reported in the FITS files will usually differ, because the program
* re-distributes the area into non-overlapping rectangular regions to simplify
* building the covariance matrix.
*
* The standard output is a list of Karhunen-Loeve modes followed by a list of
* phase screens. The format is for each mode and each screen a description of
* parameters in comments (ie, in lines starting with the sharp). Each line is
* the x and y value of the pixel in meters followed by the phase value of the mode
* in radians. Note that the modes are not yet multiplied by the (square root of) the
* eigenvalue. After each pixel row follows an empty lines. Two empty lines separate
* the modes and screens. This format is exactly the one needed to display the values
* scanning with gnuplot commands through the data,
* <tt> gnuplot</tt>
*
* <tt> gnuplot> set style data lines</tt>
*
* <tt> gnuplot> set hidden3d</tt>
*
* <tt> gnuplot> splot "out.asc" index 0</tt>
*
* <tt> gnuplot> splot "out.asc" index 1</tt>
*
* <tt> gnuplot> splot "out.asc" index 2</tt>
*
* <tt> gnuplot> quit</tt>
*
*/
int main(int argc, char *argv[])
{
	/* The command line option switch
	*/
	char oc ;

	/* Default Fried parameter of 15 cm.
	*/
	double rFried = 0.15 ;

	/* Default Kolmogorov exponent.
	*/
	double kolg_exp = 2./3. ;

	/* Default outer scale. A value <= zero is synonymous to infinity.
	*/
	double outscal = -1. ;

	/* Default width of the phase screen. 1.8 meter.
	*/
	double width = 1.8 ;

	/* Default length of the phase screen, same as the width.
	*/
	double length = width ;

	/* Default number of pixels. 100.
	*/
	int pixN = 100 ;

	/* Default number of phase screens. 100
	*/
	int screenN = 100 ;

	/* Seed of the random number generator.
	*/
	int seed = 0 ;

#ifdef HAVE_CCFITS
	/* Default output file.
	*/
	std::string fitsout ;
#endif /* HAVE_CCFITS */

#ifdef HAVE_CCFITS
        while ( (oc=getopt(argc,argv,"w:l:P:N:f:r:L:g:hs:")) != -1)
#else
        while ( (oc=getopt(argc,argv,"w:l:P:N:r:L:g:hs:")) != -1)
#endif /* HAVE_CCFITS */
        {
                switch(oc)
                {
                case 'w' :
                        width= atof(optarg) ;
                        break ;
                case 'l' :
                        length=atof(optarg) ;
                        break ;
                case 'P' :
                        pixN=atoi(optarg) ;
                        break ;
                case 'N' :
                        screenN=atoi(optarg) ;
                        break ;
                case 'r' :
                        rFried=atof(optarg) ;
                        break ;
                case 'L' :
                        outscal=atof(optarg) ;
                        break ;
                case 'g' :
                        kolg_exp=atof(optarg) ;
                        break ;
                case 'f' :
                        fitsout=optarg ;
                        break ;
                case 's' :
                        seed=atoi(optarg) ;
                        break ;
                case 'h' :
                        /* help */
                        break ;
                case '?' :
                        std::cerr << "Invalid command line option " << optopt << std::endl ;
                        return 1 ;
                }
        }

	/* TODO: check that pixN>=1 etc,
	*/

	/* check that strpcnt is >0 if this does not
	* re-read an existing fits file
	*/

	FrameSet fs ;

	/* Count of the stripes specified on the command line.
	*/
	int strpcnt = (argc-optind)/2 ;
	std::vector<double> x ;
	std::vector<double> y ;
	for(int s =0 ; s < strpcnt ; s++)
	{
		x.push_back( atof(argv[optind+2*s])) ;
		y.push_back( atof(argv[optind+2*s+1])) ;
	}

	/* area of all stripes (assumed non-overlapping...)
	*/
	double ar = x.size()*width*length ;

	double pixscal = sqrt(ar/pixN) ;

	int dimx = length/pixscal + 0.5 ;
	int dimy = width/pixscal + 0.5 ;

	/* create the grid of pixels covering all stripes.
	*/
	for(int s =0 ; s < strpcnt ; s++)
	{
		Point pt( (int)(x[s]/pixscal+0.5), (int)( y[s]/pixscal+0.5) ) ;
		Frame fr(dimx,dimy,pt) ;
		fs += fr ;
	}

	/* Assemble the turbulence parameters in a VKarCorr object.
	*/
	VKarCorr<double> vk(rFried, outscal,kolg_exp) ;

	/** Generate the modes and store them in the VKarCorr object
	*/
	vk.genmods(fs,pixscal) ;

	/** Generate the screens and store them in the VKarCorr object
	*/
	vk.genscr(screenN,seed,fs) ;

	/* print a table of x, y and mode values to stdout.
	*/
	vk.plot(std::cout,fs) ;

#ifdef HAVE_CCFITS
	CCfits::FITS::setVerboseMode(true) ;

	/* Convert the mode and screen information to a Fits table.
	*/
	CCfits::FITS *fout = vk.toFITS(fitsout,fs) ;

	/* At this point in time, the FITS file is written.
	*/
	delete fout ;
#endif /* HAVE_CCFITS */
		
	return 0 ;
}
